﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
Life#Gaze v1.4 by MacGruber.

Version 1.4 2020-12-30.
	- Fixing rotation issues when character is not upright.
	- Added option to provide a custom reference control, defaults to chestControl as before.
	- Rearanged UI and minor cleanup/fixes.

History:
	- Gaze v1.3 by MacGruber. Adding roll min/max and horizontal eye angle offset.
	- Gaze v1.1 by MacGruber. Fixes when calculating the target direction. Various improvements.
	- Gaze v1.0 by MacGruber. Just minor changes.
	- Code ported to new style VaM plugin by VeeRifter 25/10/2018
	- Original VaM ScriptEngine plugin created by MacGruber 07/09/2018

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.UI;
using System.Collections.Generic;

namespace MacGruber
{
    public class Gaze : MVRScript
    {	
        public override void Init()
		{
			if (containingAtom.type != "Person")
			{
				SetupInfoText(
					"<color=#606060><size=40><b>Gaze v1.4</b></size></color><color=#FF0000>Gaze plugin requires atom of type Person!</color>",
					300.0f, true
				);
				
				SuperController.LogError($"Gaze plugin requires atom of type Person, instead selected: {containingAtom.type}");
				return;
			}
			
			SetupInfoOneLine("<color=#606060><size=40><b>Gaze v1.4</b></size></color>", true);

			person = containingAtom;
			head = GetTransform(person, "headControl");
			eyeTarget = GetTransform(person, "eyeTargetControl");
			
			lookAtEyeTarget = SetupToggle("LookAt EyeTarget", false, true);
			gazeDuration = SetupSlider("Gaze Duration", 0.7f, 0f, 3f, true);
			
			referenceAtom = new JSONStorableStringChooser("Reference Atom", GetAllAtomUIDs(), containingAtom.uid, "Reference Atom");						
			referenceAtom.setCallbackFunction += (string v) => { UpdateReferenceControls(); };
			referenceControl = new JSONStorableStringChooser("Reference Control", new List<string>(), "", "Reference Control");
			referenceControl.setCallbackFunction += (string v) => { UpdateReferenceControls(); };
			UpdateReferenceControls();
			CreateFilterablePopup(referenceAtom, false);
			CreateFilterablePopup(referenceControl, false);
			RegisterStringChooser(referenceAtom);
			RegisterStringChooser(referenceControl);
			
			maxAngleH = SetupSlider("Max Angle Horizontal", 50.0f, 10.0f, 130.0f, false);
			maxAngleV = SetupSlider("Max Angle Vertical", 30.0f, 10.0f, 70.0f, true);
							
			focusChangeDurationMin = SetupSlider("Focus Duration Min", 1f, 1f, 10f, true);
			focusChangeDurationMax = SetupSlider("Focus Duration Max", 4f, 1f, 10f, true);
			focusChangeDurationMin.setCallbackFunction += (float dmin) => focusChangeDurationMax.val = Mathf.Max(focusChangeDurationMax.val, dmin);				
			focusChangeDurationMax.setCallbackFunction += (float dmax) => focusChangeDurationMin.val = Mathf.Min(focusChangeDurationMin.val, dmax);

			focusAngleH = SetupSlider("Focus Angle Horizontal", 4f, 1f, 15f, false);				
			focusAngleV = SetupSlider("Focus Angle Vertical", 6f, 1f, 15f, false);			
			
			rollChangeDurationMin = SetupSlider("Roll Duration Min", 2f, 1f, 10f, true);
			rollChangeDurationMax = SetupSlider("Roll Duration Max", 6f, 1f, 10f, true);
			rollChangeDurationMin.setCallbackFunction += (float rmin) => rollChangeDurationMax.val = Mathf.Max(rollChangeDurationMax.val, rmin);
			rollChangeDurationMax.setCallbackFunction += (float rmax) => rollChangeDurationMin.val = Mathf.Min(rollChangeDurationMin.val, rmax);

			rollAngleMin = SetupSlider("Roll Angle Min", -6f, -25f, 25f, false);
			rollAngleMax = SetupSlider("Roll Angle Max", 6f, -25f, 25f, false);
			
			eyeAngleVOffset = SetupSlider("Eye Angle Vertical Offset", -5.0f, -20.0f, 20.0f, false);
			eyeAngleHOffset = SetupSlider("Eye Angle Horizontal Offset", 0.0f, -20.0f, 20.0f, false);
			eyePositionOffset = SetupSlider("Eye Position Offset", 0.1f, -0.2f, 0.2f, true);
			lookAtPositionOffset = SetupSlider("LookAt Position Offset", 0.0f, -0.2f, 0.2f, true);
			
			
			SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
			SuperController.singleton.onAtomUIDsChangedHandlers += OnAtomUIDsChanged;
			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomUIDRename;
		}
		
		private void OnDestroy()
        {
			SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
			SuperController.singleton.onAtomUIDsChangedHandlers -= OnAtomUIDsChanged;
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomUIDRename;
        }
		
		private List<string> GetAllAtomUIDs()
		{
			// working around VaM weirdness not returning hidden atoms
			SuperController sc = SuperController.singleton;
			bool wasSorted = sc.sortAtomUIDs;
			sc.sortAtomUIDs = false;
			List<string> atomUIDs = sc.GetAtomUIDs();
			sc.sortAtomUIDs = wasSorted;
			return atomUIDs;
		}
				
		private void OnSceneLoaded()
		{
			OnAtomUIDsChanged(GetAllAtomUIDs());
		}
				
		private void OnAtomUIDsChanged(List<string> atomUIDs)
		{			
			atomUIDs = new List<string>(GetAllAtomUIDs());
			atomUIDs.Sort();
			referenceAtom.choices = atomUIDs;
			
			referenceAtomPrevious = referenceAtom.val; // remember atom/control in case this is a rename
			referenceControlPrevious = referenceControl.val;
			
			if (atomUIDs.Count == 0)
				referenceAtom.valNoCallback = "";
			else if (atomUIDs.Contains(referenceAtom.val))
				; // nothing to do
			else if (atomUIDs.Contains(containingAtom.uid))
				referenceAtom.valNoCallback = containingAtom.uid;
			else
				referenceAtom.valNoCallback = atomUIDs[0];
				
			UpdateReferenceControls();
		}
		
		private void OnAtomUIDRename(string oldName, string newName)
		{
			if (referenceAtomPrevious == oldName)
			{				
				referenceAtom.valNoCallback = newName;
				referenceControl.valNoCallback = referenceControlPrevious;
				UpdateReferenceControls();
			}
		}
		
		private void UpdateReferenceControls()
		{			
			Atom atom = GetAtomById(referenceAtom.val);
			List<string> controls;
			if (atom == null)
			{
				controls = new List<string>();
			}
			else
			{			
				controls = new List<string>(atom.freeControllers.Length);
				for (int i=0; i<atom.freeControllers.Length; ++i)
				{
					string storeId = atom.freeControllers[i].storeId;
					if (storeId != "headControl")
						controls.Add(storeId);
				}
			}
			referenceControl.choices = controls;
						
			if (controls.Count == 0)
				referenceControl.valNoCallback = "";
			else if (controls.Contains(referenceControl.val))
				; // nothing to do
			else if (controls.Contains("chestControl"))
				referenceControl.valNoCallback = "chestControl";
			else
				referenceControl.valNoCallback = controls[0];
			
			reference = GetTransform(atom, referenceControl.val);
		}
				
		private JSONStorableFloat SetupSlider(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}
		
		private JSONStorableBool SetupToggle(string name, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(name, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateToggle(storable, rightSide);
			RegisterBool(storable);
			return storable;
		}
		
		private JSONStorableString SetupInfoText(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
		
		public JSONStorableString SetupInfoOneLine(string text, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.UItext.alignment = TextAnchor.LowerLeft;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			textfield.height = layout.preferredHeight = layout.minHeight = 55;
			return storable;
		}

        protected void FixedUpdate()
        {		
            if (head == null || reference == null)
                return;

            // compute horizontal and vertical angles
			Transform lookAtTarget = CameraTarget.centerTarget.transform;
			if (lookAtEyeTarget.val)
				lookAtTarget = eyeTarget;
			Vector3 lookAtPosition = lookAtTarget.TransformPoint(lookAtPositionOffset.val * Vector3.up);
            Vector3 actualDir = reference.InverseTransformDirection(head.forward);
			Vector3 eyeCenterPosition = head.TransformPoint(eyePositionOffset.val * Vector3.up);
            Vector3 targetDir = lookAtPosition - eyeCenterPosition;
            targetDir.Normalize();
            targetDir = reference.InverseTransformDirection(targetDir);
            Vector2 actualDirH = new Vector2(actualDir.x, actualDir.z);
            Vector2 targetDirH = new Vector2(targetDir.x, targetDir.z);
            Vector2 actualDirV = new Vector2(actualDirH.magnitude, actualDir.y);
            Vector2 targetDirV = new Vector2(targetDirH.magnitude, targetDir.y);
            actualDirH.Normalize();
            targetDirH.Normalize();
            actualDirV.Normalize();
            targetDirV.Normalize();
            float actualH = Mathf.Atan2(actualDirH.x, actualDirH.y);
            float targetH = Mathf.Atan2(targetDirH.x, targetDirH.y);
            float actualV = Mathf.Atan2(actualDirV.y, actualDirV.x);
            float targetV = Mathf.Atan2(targetDirV.y, targetDirV.x);
			
			targetH += eyeAngleHOffset.val * Mathf.Deg2Rad;
			targetV += eyeAngleVOffset.val * Mathf.Deg2Rad;

            // apply focus
            focusChangeClock += focusChangeSpeed * Time.fixedDeltaTime;
            if (focusChangeClock >= 1.0f)
            {
                focusChangeSpeed = 1.0f / Random.Range(focusChangeDurationMin.val, focusChangeDurationMax.val);
                focusChangeClock = 0.0f;
                focusPrev = focusNext;
                focusNext = Random.insideUnitCircle;
            }
            float t = Mathf.SmoothStep(0.0f, 1.0f, focusChangeClock);
            targetH += Mathf.Lerp(focusPrev.x, focusNext.x, t) * focusAngleH.val * Mathf.Deg2Rad;
            targetV += Mathf.Lerp(focusPrev.y, focusNext.y, t) * focusAngleV.val * Mathf.Deg2Rad;

            // adjust angles
			float mah = maxAngleH.val * Mathf.Deg2Rad;
			float mav = maxAngleV.val * Mathf.Deg2Rad;
            targetH = Mathf.Clamp(targetH, -mah, mah);
            targetV = Mathf.Clamp(targetV, -mav, mav);

            actualH = Mathf.SmoothDamp(actualH, targetH, ref velocityH, gazeDuration.val, Mathf.Infinity, Time.fixedDeltaTime);
            actualV = Mathf.SmoothDamp(actualV, targetV, ref velocityV, gazeDuration.val, Mathf.Infinity, Time.fixedDeltaTime);
			
			 // apply roll
            rollChangeClock += rollChangeSpeed * Time.fixedDeltaTime;
            if (rollChangeClock >= 1.0f)
            {
                rollChangeSpeed = 1.0f / Random.Range(rollChangeDurationMin.val, rollChangeDurationMax.val);
                rollChangeClock = 0.0f;
                rollPrev = rollNext;
				float rmin = Mathf.Min(rollAngleMin.val, rollAngleMax.val);
				float rmax = Mathf.Max(rollAngleMin.val, rollAngleMax.val);
                rollNext = Random.Range(rmin, rmax) * Mathf.Deg2Rad;
            }
            t = Mathf.SmoothStep(0.0f, 1.0f, rollChangeClock);
            float roll = Mathf.Lerp(rollPrev, rollNext, t);
			Quaternion rollRot = Quaternion.Euler(0, 0, roll * Mathf.Rad2Deg);
			rollRot *= reference.rotation;

            // recombine			
            actualDir = RecombineDirection(actualH, actualV);
            actualDir = reference.TransformDirection(actualDir);
			Vector3 actualUp = rollRot * Vector3.up;
            head.transform.LookAt(head.transform.position + actualDir, actualUp);
        }

        public Transform GetTransform(Atom atom, string controlID)
        {
            return atom?.GetStorableByID(controlID)?.transform;
        }
		
        private Vector3 RecombineDirection(float angleH, float angleV)
        {
            float cosV = Mathf.Cos(angleV);
            return new Vector3(
                Mathf.Sin(angleH) * cosV,
                Mathf.Sin(angleV),
                Mathf.Cos(angleH) * cosV
            );
        }

        private Atom person;  
        private Transform head;		
        private Transform reference;
		private Transform eyeTarget;

        // tweak parameters
		private JSONStorableStringChooser referenceAtom;
		private JSONStorableStringChooser referenceControl;
		private string referenceAtomPrevious;
		private string referenceControlPrevious;
		
		
		private JSONStorableFloat maxAngleH;
		private JSONStorableFloat maxAngleV;
		private JSONStorableFloat gazeDuration;
		private JSONStorableFloat focusChangeDurationMin;
		private JSONStorableFloat focusChangeDurationMax;
		private JSONStorableFloat focusAngleV;
		private JSONStorableFloat focusAngleH;
		private JSONStorableFloat rollChangeDurationMin;
		private JSONStorableFloat rollChangeDurationMax;
		private JSONStorableFloat rollAngleMin;
		private JSONStorableFloat rollAngleMax;		
		private JSONStorableFloat eyePositionOffset;
		private JSONStorableFloat eyeAngleVOffset;
		private JSONStorableFloat lookAtPositionOffset;
		private JSONStorableFloat eyeAngleHOffset;
		private JSONStorableBool lookAtEyeTarget;

        // runtime data
        private float velocityH = 0.0f;
        private float velocityV = 0.0f;
        private float focusChangeClock = 1.0f;
        private float focusChangeSpeed = 1.0f;
        private Vector2 focusNext = Vector2.zero;
        private Vector2 focusPrev = Vector2.zero;
        private float rollNext = 0.0f;
        private float rollPrev = 0.0f;
        private float rollChangeClock = 1.0f;
        private float rollChangeSpeed = 1.0f;
    }
}
